string SufamiTurbo = R"(

database revision=2013-01-22

release
  cartridge linkable
    rom name=program.rom size=0x80000
    ram name=save.ram size=0x800
  information
    title:    SDウルトラバトル ウルトラマン伝説
    name:     SD Ultra Battle - Ultraman Densetsu
    region:   JP
    revision: 1.0
    serial:   SFT-0101-JPN
    sha256:   2bb55214fb668ca603d7b944b14f105dfb10b987a8902d420fe4ae1cb69c1d4a
    configuration
      rom name=program.rom size=0x80000
      ram name=save.ram size=0x800
      linkable

release
  cartridge linkable
    rom name=program.rom size=0x80000
    ram name=save.ram size=0x800
  information
    title:    SDウルトラバトル セブン伝説
    name:     SD Ultra Battle - Seven Densetsu
    region:   JP
    revision: 1.0
    serial:   SFT-0102-JPN
    sha256:   2fec5f2bc7dee010af10569a3d2bc18715a79a126940800c3eade5abbd625e3f
    configuration
      rom name=program.rom size=0x80000
      ram name=save.ram size=0x800
      linkable

release
  cartridge linkable
    rom name=program.rom size=0x80000
    ram name=save.ram size=0x800
  information
    title:    ポイポイ忍者ワールド
    name:     Poi Poi Ninja World
    region:   JP
    revision: 1.0
    serial:   SFT-0103-JPN
    sha256:   602b20b788640f5743487108a10f3f77bca5ce2d24208b25b1ca498a96eb0d69
    configuration
      rom name=program.rom size=0x80000
      ram name=save.ram size=0x800
      linkable

release
  cartridge linkable
    rom name=program.rom size=0x80000
    ram name=save.ram size=0x2000
  information
    title:    SDガンダムジェネレーション 一年戦争記
    name:     SD Gundam Generation - Ichinen Sensouki
    region:   JP
    revision: 1.0
    serial:   SFT-0104-JPN
    sha256:   3e82215bed08274874b30d461fc4a965c6bca932229da5d46d56e36f484d65eb
    configuration
      rom name=program.rom size=0x80000
      ram name=save.ram size=0x2000
      linkable

release
  cartridge linkable
    rom name=program.rom size=0x80000
    ram name=save.ram size=0x2000
  information
    title:    SDガンダムジェネレーション グリプス戦記
    name:     SD Gundam Generation - Grips Senki
    region:   JP
    revision: 1.0
    serial:   SFT-0105-JPN
    sha256:   8547a08ed11fe408eac282a90ac46654bd2e5f49bda3aec8e5edf166a0a4b9af
    configuration
      rom name=program.rom size=0x80000
      ram name=save.ram size=0x2000
      linkable

release
  cartridge
    rom name=program.rom size=0x80000
  information
    title:    ゲゲゲの鬼太郎 妖怪ドンジャラ
    name:     Gegege no Kitarou - Youkai Donjara
    region:   JP
    revision: 1.0
    serial:   SFT-0106-JPN
    sha256:   d93b3a570e7cf343f680ab0768a50b77e3577f9c555007e2de3decd6bc4765c8
    configuration
      rom name=program.rom size=0x80000

release
  cartridge linkable
    rom name=program.rom size=0x80000
    ram name=save.ram size=0x2000
  information
    title:    SDガンダムジェネレーション アクシズ戦記
    name:     SD Gundam Generation - Axis Senki
    region:   JP
    revision: 1.0
    serial:   SFT-0107-JPN
    sha256:   2a9d7c9a61318861028a73ca03e32a48cff162d76cba36fbaab8690b212efe9b
    configuration
      rom name=program.rom size=0x80000
      ram name=save.ram size=0x2000
      linkable

release
  cartridge linkable
    rom name=program.rom size=0x80000
    ram name=save.ram size=0x2000
  information
    title:    SDガンダムジェネレーション バビロニア建国戦記
    name:     SD Gundam Generation - Babylonia Kenkoku Senki
    region:   JP
    revision: 1.0
    serial:   SFT-0108-JPN
    sha256:   60ac017c18f534e8cf24ca7f38e22ce92db95ea6c30b2d59d76f13c4f1c8a6e4
    configuration
      rom name=program.rom size=0x80000
      ram name=save.ram size=0x2000
      linkable

release
  cartridge linkable
    rom name=program.rom size=0x80000
    ram name=save.ram size=0x2000
  information
    title:    SDガンダムジェネレーション ザンスカール戦記
    name:     SD Gundam Generation - Zanscar Senki
    region:   JP
    revision: 1.0
    serial:   SFT-0110-JPN
    sha256:   5951a58a91d8e397d0a237ccc2b1248e17c7312cb9cc11cbc350200a97b4e021
    configuration
      rom name=program.rom size=0x80000
      ram name=save.ram size=0x2000
      linkable

release
  cartridge linkable
    rom name=program.rom size=0x80000
    ram name=save.ram size=0x800
  information
    title:    SDガンダムジェネレーション コロニー格闘記
    name:     SD Gundam Generation - Colony Kakutouki
    region:   JP
    revision: 1.0
    serial:   SFT-0111-JPN
    sha256:   e639b5d5d722432b6809ccc6801dc584e1a3016379f34b335ed2dfa73b1ebf69
    configuration
      rom name=program.rom size=0x80000
      ram name=save.ram size=0x800
      linkable

)";
